@@ -4,10 +4,12 @@ import xgboost as xgb
 import lightgbm as lgb
 import pandas as pd
 import time
-#import category_encoders as ce
+import category_encoders as ce
 from hyperopt import hp, Trials, STATUS_OK, fmin, tpe, anneal
 import matplotlib.pyplot as plt
 from statsmodels.nonparametric.smoothers_lowess import lowess
+import json
+
 
 def _get_best_params(best, params, int_params, trials):
     best_params = best.copy()
@@ -23,7 +25,7 @@ def _get_best_params(best, params, int_params, trials):
 
 def _evaluation_plot(losses, best='min', invert=False, figsize=(15,6)):
 
-    if invert == True:
+    if invert:
         losses = [ x * -1 for x in losses ]
     x = np.arange(0, len(losses))
     ys = lowess(losses, x)[:, 1]
@@ -54,6 +56,13 @@ def _update_params(hyperopt_params, params):
     return hyperopt_params
 
 
+def _save_json(data, path):
+    if path[path.rfind('.'):] != '.json':
+        path = path + '.json'
+    with open(path, 'w') as fp:
+        json.dump(data, fp)
+
+
 def opt_catboost(X_train,
                  y_train,
                  cat_features=None,
@@ -66,7 +75,8 @@ def opt_catboost(X_train,
                  max_rounds=5000,
                  seed=0,
                  early_stopping=20,
-                 figsize=(15,6)):
+                 figsize=(15,6),
+                 savepath=None):
     """
     Performs n_trials using Hyperopt to obtain the best catboost parameters
 
@@ -87,14 +97,14 @@ def opt_catboost(X_train,
     """
     HYPEROPT_PARAMS = {'learning_rate': hp.loguniform('learning_rate', np.log(0.01), np.log(0.3)),
                        'random_state': hp.choice('random_state', [0, 1, 2, 3]),
-                       'l2_leaf_reg': hp.loguniform('l2_leaf_reg', 0, np.log(20)), #reg_lambda
+                       'l2_leaf_reg': hp.loguniform('l2_leaf_reg', 0, np.log(20)),  # eg_lambda
                        'bagging_temperature': hp.uniform('bagging_temperature', 0, 1),
                        'random_strength': hp.uniform('random_strength', 0.5, 3),
                        'depth': hp.quniform('depth', 2, 8, 2),  # 10,
-                       'rsm': hp.uniform('rsm', 0.1, 0.9) if device != 'GPU' else None, #colsample_bylevel
+                       'rsm': hp.uniform('rsm', 0.1, 0.9) if device != 'GPU' else None,  # colsample_bylevel
                        'loss_function': 'RMSE',
                        'eval_metric': 'RMSE',
-                       'max_bin': hp.qloguniform('max_bin', np.log(32), np.log(64), 5), #border count
+                       'max_bin': hp.qloguniform('max_bin', np.log(32), np.log(64), 5),  # border count
                        'od_type': 'Iter',  # 'IncToDec',  # 'Iter'
                        'od_wait': early_stopping,
                        'task_type': device # GPU devices
@@ -108,6 +118,10 @@ def opt_catboost(X_train,
     # Catboost dataset
     train_pool = cb.Pool(data=X_train, label=y_train, cat_features=cat_features)
 
+    # If folds already provided, set cv_folds to None
+    if folds:
+        cv_folds = None
+
     # Objective function for Hyperopt
     def objective(params):
         start = time.time()
@@ -137,7 +151,7 @@ def opt_catboost(X_train,
 
         if verbose > 0:
             print('Train Loss: %0.4f, Test Loss: %0.4f RMSE with %d iterations. Time elapsed %s' % (
-            train_loss_value, test_loss_value, best_iteration, str(round(time.time() - start,2))))
+                train_loss_value, test_loss_value, best_iteration, str(round(time.time() - start,2))))
 
         val = {'loss': test_loss_value,  # mandatory
                'status': STATUS_OK,  # mandatory
@@ -159,7 +173,12 @@ def opt_catboost(X_train,
     if verbose == 2:
         _evaluation_plot(trials.losses(), figsize=figsize)
 
-    return {'params':best_params, 'n':best_n}
+    best = {'params': best_params, 'n': best_n}
+
+    if savepath:
+        _save_json(best , savepath)
+
+    return best
 
 
 def opt_xgboost(X_train,
@@ -174,13 +193,15 @@ def opt_xgboost(X_train,
                 max_rounds=5000,
                 seed=0,
                 early_stopping=20,
-                figsize=(15,6)):
+                figsize=(15, 6),
+                savepath=None):
     """
     Performs n_trials using Hyperopt to obtain the best XGBoost parameters
 
     Inputs:
         X_train, y_train, cat_features: Data to build Pool object
         cv_folds: number of CV folds for validation on each trial
+        folds: Custom splitting indices. This parameter has the highest priority among other data split parameters.
         n_trials: number of trials to perform
         verbose: 0 = no log and no plot, 1 = log but no plot, 2 = log and plot
         params: dict to override the Hyperopt params
@@ -214,7 +235,7 @@ def opt_xgboost(X_train,
 
     # Transform categorical features to one_hot
     if cat_features:
-        X_train = ce.one_hot.OneHotEncoder(cols = cat_features, drop_invariant=True).fit_transform(X_train)
+        X_train = ce.one_hot.OneHotEncoder(cols=cat_features, drop_invariant=True).fit_transform(X_train)
     train_dmatrix = xgb.DMatrix(X_train, y_train)
 
     def objective(params):
@@ -244,21 +265,21 @@ def opt_xgboost(X_train,
 
         if verbose > 0:
             print('Train Loss: %0.4f, Test Loss: %0.4f RMSE with %d iterations. Time elapsed %s' % (
-            train_loss_value, test_loss_value, best_iteration, str(round(time.time() - start,2))))
+                train_loss_value, test_loss_value, best_iteration, str(round(time.time() - start,2))))
 
         val = {'loss': test_loss_value,  # mandatory
-                'status': STATUS_OK,  # mandatory
-                'best_n_iters': best_iteration}
+               'status': STATUS_OK,  # mandatory
+               'best_n_iters': best_iteration}
 
         return val
 
     trials = Trials()
     hyperopt_xgb = fmin(fn=objective,
-                       space=HYPEROPT_PARAMS,
-                       algo=tpe.suggest,
-                       verbose=True if verbose > 0 else False,
-                       max_evals=n_trials,
-                       trials=trials)
+                        space=HYPEROPT_PARAMS,
+                        algo=tpe.suggest,
+                        verbose=True if verbose > 0 else False,
+                        max_evals=n_trials,
+                        trials=trials)
 
     best_params, best_n = _get_best_params(hyperopt_xgb, HYPEROPT_PARAMS, INTEGER_PARAMS_LIST, trials)
 
@@ -266,7 +287,12 @@ def opt_xgboost(X_train,
     if verbose == 2:
         _evaluation_plot(trials.losses(), figsize=figsize)
 
-    return {'params': best_params, 'n': best_n}
+    best = {'params': best_params, 'n': best_n}
+
+    if savepath:
+        _save_json(best, savepath)
+
+    return best
 
 
 def opt_lightgbm(X_train,
@@ -281,13 +307,15 @@ def opt_lightgbm(X_train,
                  max_rounds=5000,
                  seed=0,
                  early_stopping=20,
-                 figsize=(15,6)):
+                 figsize=(15, 6),
+                 savepath=None):
     """
     Performs n_trials using Hyperopt to obtain the best LightGBM parameters
 
     Inputs:
         X_train, y_train, cat_features: Data to build Pool object
         cv_folds: number of CV folds for validation on each trial
+        folds: Custom splitting indices. This parameter has the highest priority among other data split parameters.
         n_trials: number of trials to perform
         verbose: 0 = no log and no plot, 1 = log but no plot, 2 = log and plot
         params: dict to override the Hyperopt params
@@ -300,22 +328,22 @@ def opt_lightgbm(X_train,
         dict that contains the best params and the best number of iterations
     """
     HYPEROPT_PARAMS = {'learning_rate': hp.loguniform('learning_rate', np.log(0.01), np.log(0.3)),
-                        'num_leaves': hp.qloguniform('num_leaves', np.log(32), np.log(256), 5),
-                        'max_depth': hp.quniform('max_depth', 2, 10, 2),  # 10,
-                        'min_data_in_leaf': hp.qloguniform('min_data_in_leaf', np.log(32), np.log(256), 5),
-                        # 'min_child_weight': hp.loguniform('min_child_weight', -16, 5)
-                        'bagging_fraction': hp.uniform('bagging_fraction', 0.2, 0.9),
-                        'bagging_freq': hp.uniform('bagging_freq', 1, 10),
-                        'feature_fraction': hp.uniform('feature_fraction', 0.2, 0.9),
-                        'lambda_l1': hp.loguniform('lambda_l1', 0, np.log(100)),
-                        'lambda_l2': hp.loguniform('lambda_l2', 0, np.log(20)),
-                        #'max_bin': hp.qloguniform('max_bin', np.log(32), np.log(64), 5), #border count,
-                        'metric': 'rmse',
-                        'seed': hp.choice('seed', [0, 1, 2, 3]),
-                        'num_threads': 5,
-                        # 'gpu_id': 0,
-                        'device_type':device.lower()
-                        }
+                       'num_leaves': hp.qloguniform('num_leaves', np.log(32), np.log(256), 5),
+                       'max_depth': hp.quniform('max_depth', 2, 10, 2),  # 10,
+                       'min_data_in_leaf': hp.qloguniform('min_data_in_leaf', np.log(32), np.log(256), 5),
+                       # 'min_child_weight': hp.loguniform('min_child_weight', -16, 5)
+                       'bagging_fraction': hp.uniform('bagging_fraction', 0.2, 0.9),
+                       'bagging_freq': hp.uniform('bagging_freq', 1, 10),
+                       'feature_fraction': hp.uniform('feature_fraction', 0.2, 0.9),
+                       'lambda_l1': hp.loguniform('lambda_l1', 0, np.log(100)),
+                       'lambda_l2': hp.loguniform('lambda_l2', 0, np.log(20)),
+                       # 'max_bin': hp.qloguniform('max_bin', np.log(32), np.log(64), 5), # border count,
+                       'metric': 'rmse',
+                       'seed': hp.choice('seed', [0, 1, 2, 3]),
+                       'num_threads': 5,
+                       # 'gpu_id': 0,
+                       'device_type':device.lower()
+                       }
 
     HYPEROPT_PARAMS = _update_params(HYPEROPT_PARAMS, params)
     INTEGER_PARAMS_LIST = ['max_depth', 'seed', 'bagging_freq','num_leaves','min_data_in_leaf']
@@ -347,7 +375,7 @@ def opt_lightgbm(X_train,
 
         if verbose > 0:
             print('Test Loss: %0.4f RMSE with %d iterations. Time elapsed %s' % (
-            loss, best_iteration, str(round(time.time() - start, 2))))
+                loss, best_iteration, str(round(time.time() - start, 2))))
 
         val = {'loss': loss,  # mandatory
                'status': STATUS_OK,  # mandatory
@@ -365,12 +393,17 @@ def opt_lightgbm(X_train,
 
     best_params, best_n = _get_best_params(hyperopt_lgb, HYPEROPT_PARAMS, INTEGER_PARAMS_LIST, trials)
     # apparently hyperopt has a bug on choice methods
-    #best_params['num_leaves'] = 2 ** (best_params['num_leaves']+3) - 1
-    #best_params['min_data_in_leaf'] = 2 ** (best_params['min_data_in_leaf']+3) - 1
-    #best_params['max_bin'] = 2 ** (best_params['max_bin']+3) - 1
+    # best_params['num_leaves'] = 2 ** (best_params['num_leaves']+3) - 1
+    # best_params['min_data_in_leaf'] = 2 ** (best_params['min_data_in_leaf']+3) - 1
+    # best_params['max_bin'] = 2 ** (best_params['max_bin']+3) - 1
 
     # Plot results
     if verbose == 2:
         _evaluation_plot(trials.losses(), figsize=figsize)
 
-    return {'params': best_params, 'n': best_n}
+    best = {'params': best_params, 'n': best_n}
+
+    if savepath:
+        _save_json(best, savepath)
+
+    return best
